searchData={"content_type":"text/markdown","items":[{"doc":"Testing with side-effects is often hard. Various solutions exist to work around\nthe difficulties, e.g. mocking. This library offers a very easy way to achieve\ntestable code by mocking. Instead of mocking we talk about binding effects to another implementation.\n`Efx` offers a declarative way to mark effectful functions and bind them in tests.\n\nEfx allows async testing even in with child-processes, since it uses process-dictionaries\nto store bindings and find them in the supervision-tree (see this [test-case](https://github.com/bravobike/efx/blob/improve-doc-example/test/efx_case_test.exs#L52)).","ref":"Efx.html","title":"Efx","type":"module"},{"doc":"Efx is a small library that does one thing and one thing only very well: Make code\nthat contains side effects testable.\n\nExisting mock libraries often set up mocks in non-declarative ways: configs need\nto be adapted & mock need to be initialized. In source code there are intrusive\ninstructions to set up mockable code. `Efx` is very unintrusive in both, source\ncode and test code. It offers a convenient and declarative syntax. Instead of\nmocking we talk about binding effects.\n\nEfx follows the following principles:\n\n- Implementing and binding effects should be as simple and declarative as possible.\n- Modules contain groups of effects that can only be bound as a set.\n- We want to run as many tests async as possible. Thus, we traverse\nthe supervision tree to find rebound effects in the spawning test processes,\nin an isolated manner.\n- Effects by default execute their default implementation in tests, and thus, must be explicitly bound.\n- Effects can only be bound in tests, but not in production. In production, the default implementation is always executed.\n- We want zero performance overhead in production.","ref":"Efx.html#module-rationale","title":"Rationale - Efx","type":"module"},{"doc":"","ref":"Efx.html#module-usage","title":"Usage - Efx","type":"module"},{"doc":"Given the following code:\n\n    defmodule MyModule do\n\n      def read_data() do\n        File.read!(\"file.txt\")\n        |> deserialize()\n      end\n\n      def write_data(data) do\n        serialized_data = data |> serialize()\n        File.write!(\"file.txt\", deserialized_data)\n      end\n\n      defp deserialize(raw) do\n        ...\n      end\n\n      defp serialize(data) do\n        ...\n      end\n\n    end\n\nIn this example, it's quite complicated to test deserialization and serialization since\nwe have to prepare and place the file correctly for each test.\n\nWe can rewrite the module using `Efx` as follows:\n\n\n    defmodule MyModule do\n\n      use Efx\n\n      def read_data() do\n        read_file!()\n        |> deserialize()\n      end\n\n      def write_data(data) do\n        data\n        |> serialize()\n        |> write_file!()\n      end\n\n      @spec read_file!() :: binary()\n      defeffect read_file!() do\n        File.read!(\"file.txt\")\n      end\n\n      @spec write_file!(binary()) :: :ok\n      defeffect write_file!(raw) do\n        File.write!(\"file.txt\", raw)\n      end\n\n      ...\n\n    end\n\nBy using the `defeffect`-macro, we define an effect-function as well as provide\na default-implementation in its body. It is mandatory for each of the effect-functions to have a matching spec.\n\nThe above code is now easily testable since we can rebind the effect-functions with ease:\n\n    defmodule MyModuleTest do\n\n      use EfxCase\n\n      describe \"read_data/0\" do\n        test \"works as expected with empty file\" do\n          bind(MyModule, :read_file!, fn -> \"\" end)\n          bind(MyModule, :write_file!, fn _ -> :ok end)\n\n          # test code here\n          ...\n        end\n\n        test \"works as expected with proper contents\" do\n          bind(MyModule, :read_file!, fn -> \"some expected file content\" end)\n          bind(MyModule, :write_file!, fn _ -> :ok end)\n\n          # test code here\n          ...\n        end\n\n      end\n\n    end\n\nInstead of returning the value of the default implementation, `MyModule.read_file!/0` returns test data that is needed for the test case. `MyModule.write_file!` does nothing.\n\nFor more details, see the `EfxCase`-module.\n\nNote that Efx generates and implements a behavior. Thus, it is recommended, to move side effects to a dedicated submodule, e.g. MyModule.Effects, to not accidentally interfere with existing behaviors.","ref":"Efx.html#module-example","title":"Example - Efx","type":"module"},{"doc":"","ref":"Efx.html#already_exists?/3","title":"Efx.already_exists?/3","type":"function"},{"doc":"","ref":"Efx.html#defeffect/2","title":"Efx.defeffect/2","type":"macro"},{"doc":"","ref":"Efx.Application.html","title":"Efx.Application","type":"module"},{"doc":"","ref":"Efx.Application.html#start/2","title":"Efx.Application.start/2","type":"function"},{"doc":"Module for testing with effects.\n\nBinding effects in tests follows these principles:\n\n- By default, all effects of a module are bound to the default implementation.\n- We either bind all effect functions of a module or none.\n  We cannot bind single functions (except the explicit use of :default).\n  If we rebind only one effect and the other is called, we raise.\n- A function is either bound without or with a specified number of expected calls.\n  If a function has multiple binds, they are called in the given order, until they satisfied their\n  expected number of calls.\n- The number of expected calls is always verified.","ref":"EfxCase.html","title":"EfxCase","type":"module"},{"doc":"To bind effects, one simply has to use this module and call\nthe bind macro. Let's say we have the following effects implementation:\n\n    defmodule MyModule do\n\n      use Efx\n\n      def read_data() do\n        read_file!()\n        |> deserialize()\n      end\n\n      def write_data(data) do\n        data\n        |> serialize()\n        |> write_file!()\n      end\n\n      @spec read_file!() :: binary()\n      defeffect read_file!() do\n        File.read!(\"file.txt\")\n      end\n\n      @spec write_file!(binary()) :: :ok\n      defeffect write_file!(raw) do\n        File.write!(\"file.txt\", raw)\n      end\n\n      ...\n\n    end\n\nWe can now rebind the code in tests for different test scenarios.\n\n    defmodule MyModuleTest do\n      use EfxCase\n\n      test \"works as expected with empty file\" do\n        bind(MyModule, :read_file!, fn -> \"\" end)\n\n        # test code here\n        ...\n      end\n    end\n\nInstead of returning the value of the default implementation, `MyModule.read_file!/0` returns an empty string now, representing an empty file.\n\nOnce we bind one effect of a module, all effect-function of that module need to be bound. If an unbound effect-function is called, an error is raised.","ref":"EfxCase.html#module-binding-effects","title":"Binding effects - EfxCase","type":"module"},{"doc":"We can additionally define an expected number of calls. The expected\nnumber of calls is always verified - a test run will fail if\nit is not satisfied, as well as exceeded.\n\nWe can define a number of expected calls as follows:\n\n    test \"works as expected with empty file\" do\n      bind(MyModule, :read_file!, fn -> \"\" end, calls: 1)\n\n      # test code here\n      ...\n    end\n\nIn this case, we verify that the bound function `get/0` is called\nexactly twice.\n\nWe can also use multiple binding with the call option set. Then, they\nare executed until the number of bindings is satisfied, followed by the next\nbinding:\n\n\n    test \"works as expected with empty file and then some data\" do\n      bind(MyModule, :read_file!, fn -> \"\" end, calls: 1)\n      bind(MyModule, :read_file!, fn -> \"some meaningful data\" end, calls: 1)\n\n      # test code here\n      ...\n    end\n\nIn the above example, the test only succeeds if there are 2 calls to the effect-function.\nThe first call returns an empty string, while the second returns `\"some meaningful data\"`.","ref":"EfxCase.html#module-binding-with-an-expected-number-of-calls","title":"Binding with an expected Number of Calls - EfxCase","type":"module"},{"doc":"Effect binding uses process dictionaries to find the right binding\nby traversing the supervision-tree towards the root.\nAs long as calling processes have the testing process that defines\nthe binding as an ancestor, binding works. If we cannot ensure that,\nwe can set binding to global. However, then tests must be explicitly set\nto async to not interfere:\n\n    defmodule MyModuleTest do\n      use EfxCase, async: false\n\n      test \"async test works as expected\" do\n        bind(MyModule, :read_file!, fn -> \"\" end)\n\n        # test code here\n        ...\n      end\n    end","ref":"EfxCase.html#module-binding-globally","title":"Binding globally - EfxCase","type":"module"},{"doc":"If we want to set up the same binding for multiple tests, we can do\nthis as follows:\n\n    defmodule MyModuleTest do\n      use EfxCase, async: false\n\n      setup_effects(MyModule,\n         read_file!: fn -> \"some meaningful data\" end\n      )\n\n      test \"works with meaningful data\" do\n        # test code here\n        ...\n      end\n\n      test \"another test works with meaningful data\" do\n        # test code here\n        ...\n      end\n    end","ref":"EfxCase.html#module-setup-for-many-tests","title":"Setup for many Tests - EfxCase","type":"module"},{"doc":"While it is best practice to bind all function of a module or none,\nwe can also default certain functions explicitly:\n\n\n    defmodule MyModuleTest do\n      use EfxCase, async: false\n\n      test \"works with meaningful data\" do\n        bind(MyModule, :read_file!, fn -> \"some meaningful data\" end)\n        bind(MyModule, :write_file!, {:default, 1})\n\n        # test code here\n        ...\n      end\n    end\n\nWhile entirely leaving out `write_file!/1` would result in an error\n(when called), we can tell `Efx` to use it's\ndefault implementation. Note that when default, we have to provide the arity of the function.\nIt can be combined with an expected number of calls.","ref":"EfxCase.html#module-explicitly-defaulting-one-function-in-tests","title":"Explicitly defaulting one Function in Tests - EfxCase","type":"module"},{"doc":"","ref":"EfxCase.html#bind/5","title":"EfxCase.bind/5","type":"function"},{"doc":"","ref":"EfxCase.html#setup_effects/2","title":"EfxCase.setup_effects/2","type":"macro"},{"doc":"","ref":"EfxCase.Internal.html","title":"EfxCase.Internal","type":"module"},{"doc":"","ref":"EfxCase.Internal.html#init/1","title":"EfxCase.Internal.init/1","type":"function"},{"doc":"","ref":"EfxCase.Internal.html#verify_mocks!/1","title":"EfxCase.Internal.verify_mocks!/1","type":"function"},{"doc":"Internal logic of a mocked behaviour. This module is meant\nto make handling a mock, counting the calls, finding the\nright mocked function, calling the mocked function, etc... more\nconvenient.\n\nA mock consists of the following:\n\n- a list of mocked functions","ref":"EfxCase.Mock.html","title":"EfxCase.Mock","type":"module"},{"doc":"","ref":"EfxCase.Mock.html#add_fun/5","title":"EfxCase.Mock.add_fun/5","type":"function"},{"doc":"","ref":"EfxCase.Mock.html#get_fun/3","title":"EfxCase.Mock.get_fun/3","type":"function"},{"doc":"","ref":"EfxCase.Mock.html#get_unsatisfied/1","title":"EfxCase.Mock.get_unsatisfied/1","type":"function"},{"doc":"","ref":"EfxCase.Mock.html#inc_called/3","title":"EfxCase.Mock.inc_called/3","type":"function"},{"doc":"","ref":"EfxCase.Mock.html#make/1","title":"EfxCase.Mock.make/1","type":"function"},{"doc":"","ref":"EfxCase.Mock.html#t:t/0","title":"EfxCase.Mock.t/0","type":"type"},{"doc":"Internal logic of a mocked function. \n\nA mocked functions consists of the following:\n\n- the name or identifier of the function\n- the arity of the function\n- the implementation/replacement of the function. Besides an\n  anonymous function this can be :unmocked or :default.\n  :unmocked says there is no replacement yet and error when called \n  :default refers to the default implementation\n- a number of expected calls\n- a counter showing how the mocked functions has been called","ref":"EfxCase.Mock.MockedFun.html","title":"EfxCase.Mock.MockedFun","type":"module"},{"doc":"","ref":"EfxCase.Mock.MockedFun.html#limit_reached?/1","title":"EfxCase.Mock.MockedFun.limit_reached?/1","type":"function"},{"doc":"","ref":"EfxCase.Mock.MockedFun.html#satisfied?/1","title":"EfxCase.Mock.MockedFun.satisfied?/1","type":"function"},{"doc":"","ref":"EfxCase.Mock.MockedFun.html#t:t/0","title":"EfxCase.Mock.MockedFun.t/0","type":"type"},{"doc":"This module implements a global mutable state needed for mocking.\n\nMocking works by adding the pid of the mock-owning process to it's\ndictionary and accessing this pid using a lookup with `ProcessTree`.\n\nThe pid then can be used to lookup information of the mock, e.g. which\nfunction is mocked with replacement or how often a function was\ncalled to verify. This module is the global lookup.\n\nWe need a global lookup in which we can manipulate data to count mock\ninvocations as well as store mocks as global when a test is flagged as\nasync.","ref":"EfxCase.MockState.html","title":"EfxCase.MockState","type":"module"},{"doc":"","ref":"EfxCase.MockState.html#add_fun/5","title":"EfxCase.MockState.add_fun/5","type":"function"},{"doc":"","ref":"EfxCase.MockState.html#add_fun/6","title":"EfxCase.MockState.add_fun/6","type":"function"},{"doc":"","ref":"EfxCase.MockState.html#call/3","title":"EfxCase.MockState.call/3","type":"function"},{"doc":"","ref":"EfxCase.MockState.html#call/4","title":"EfxCase.MockState.call/4","type":"function"},{"doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"EfxCase.MockState.html#child_spec/1","title":"EfxCase.MockState.child_spec/1","type":"function"},{"doc":"","ref":"EfxCase.MockState.html#clean_globals/0","title":"EfxCase.MockState.clean_globals/0","type":"function"},{"doc":"","ref":"EfxCase.MockState.html#mocked?/1","title":"EfxCase.MockState.mocked?/1","type":"function"},{"doc":"","ref":"EfxCase.MockState.html#mocked?/2","title":"EfxCase.MockState.mocked?/2","type":"function"},{"doc":"","ref":"EfxCase.MockState.html#parse/1","title":"EfxCase.MockState.parse/1","type":"function"},{"doc":"","ref":"EfxCase.MockState.html#start_link/1","title":"EfxCase.MockState.start_link/1","type":"function"},{"doc":"","ref":"EfxCase.MockState.html#verify_called!/0","title":"EfxCase.MockState.verify_called!/0","type":"function"},{"doc":"","ref":"EfxCase.MockState.html#verify_called!/1","title":"EfxCase.MockState.verify_called!/1","type":"function"},{"doc":"","ref":"EfxCase.MockState.html#t:pid_t/0","title":"EfxCase.MockState.pid_t/0","type":"type"}],"producer":{"name":"ex_doc","version":[48,46,51,51,46,48]}}